/*
    csyobjec.prg - Class(y) Object class

    Copyright (c) 1992-1994 Anton van Straaten

    All classes ultimately inherit from the Object class.  If no
    superclass is specified when creating a class, Object is used
    automatically.

    The Object class defines methods which are essential to the
    correct functioning of the object system.  These include 'classH',
    'className', as well as a number of error handling methods.

    The init() message is mapped to the null() method, which does nothing.  
    If a class or its superclasses do not implement their own init() method, 
    this one will ultimately be invoked when Class's new() method sends an 
    init() message to the newly created instance.

    Mon  01-13-1992  11:18:49   avs - creation
    Thu  03-26-1992  12:46:55   avs
    Fri  06-19-1992  21:20:54   avs
	Sun  06-12-1994  23:35:17	avs - changes for v2.5
*/

/*
    The following declaration forces linkage of a module which supports the
    Clipper debugger's standard object inspector's behavior.  If you plan to
    always use Class(y)'s enhanced object inspector, you can comment out this
    line.
*/
EXTERNAL CsyOldInsp


#include "class(y).ch"
#include "csyprim.ch"


CREATE CLASS Object FROM CsyNoClass METACLASS Class
EXPORT:
    MESSAGE init        METHOD CsyNull

    // methods to return string representation of object.  asExpStr is
    // dynamically mapped to asString so subclasses don't need to define both
    // unless asExpStr result differs from asString, as in the scalar classes.
    METHOD  asString
    METHOD  asExpStr

    METHOD  basicSize
    METHOD  become

    // classH and className are required by Clipper.  They are also
    // referred to by the Class(y) kernel, so the methods must be public.
    MESSAGE classH      METHOD _csyClassH
    MESSAGE className   METHOD _csyClsNam

    METHOD  isEqual,	;
		    isKindOf,	;
		    isScalar,	;
		    copy,		;					// shallow copy
		    deepCopy,	;
		    deferred

    MESSAGE eval    METHOD CsyNull
    MESSAGE exec    METHOD CsyNull

    MESSAGE error   METHOD CsyError

    MESSAGE hash    METHOD deferred
    MESSAGE null    METHOD CsyNull

    MESSAGE size    METHOD basicSize    

    METHOD  protectErr,	 ;
	    	hiddenErr,	 ;
			msgNotFound, ;
			readOnlyErr, ;
			wrongClass,	 ;
			badMethod

	MESSAGE class 	PRIMITIVE CMS_CLASS ACCESS

	MESSAGE super 	PRIMITIVE CMS_SUPER, ;
			perform PRIMITIVE CMS_PERFORM

END CLASS


METHOD asString
RETURN ::class:name + " object"


/*
    asExpStr()

    Map to asString dynamically, to allow asString to be overridden without
    necessarily having to override asExpStr.
*/

METHOD asExpStr
RETURN ::asString


METHOD basicSize
RETURN LEN( self )


METHOD become( newSelf )
    ASIZE( self, LEN( newSelf ) )
    ACOPY( newSelf, self )
    CsySetDict( self, newSelf:classH )
RETURN self


FUNCTION _csyClsNam
RETURN QSELF():class:name


/*
    classH()

    This method will not be affected by reference objects, which means it
    will not allow hierarchy navigation in the debugger's native object
    inspector.  A C routine has been used instead.

METHOD classH
RETURN ::class:handle
*/


/*
    copy()

    Return a copy of the receiver which shares the receiver's instance
    variables.
*/

METHOD copy
RETURN ACOPY( self, ::class:basicNew() )


/*
    deepCopy()

    Return a copy of the receiver with shallow copies of each instance
    variable.  This is not as 'deep' as the kind of copy done by ACLONE(),
    but ACLONE() cannot handle self-referencing objects or arrays.
*/

METHOD deepCopy
    LOCAL i
    LOCAL nLen   := LEN(self)
    LOCAL newObj := ::class:basicNew()

    FOR i := 1 TO nLen
        newObj[i] := self[i]:copy()
    NEXT
RETURN newObj


METHOD isEqual( o )
RETURN self == o


/*
    isKindOf() could normally be implemented by delegation in the class
    declaration, since all it does is pass a message through to the
    object's class.  However, the Object class is created too early on
    in the Class(y) initialization process to be able to use delegation.
*/

METHOD isKindOf( oClass )
RETURN ::class:inheritsFrom( oClass )


// ScalarObject class overrides this
METHOD isScalar
RETURN .F.


/*
    :null() (CsyNull())

    Do nothing.  This method can be mapped to from subclasses, if required,
    using a command of the form:

        MESSAGE foo IS NULL

    Note we cheat by not using the METHOD command, so we don't get the line
    'self := qself()', which would be useless in this case.  It is global
    to allow use by other Class(y) system classes.
*/

FUNCTION CsyNull
RETURN NIL


#include "error.ch"
#include "csyerror.ch"

PROCEDURE CsyError( nSubCode, cDesc, cClassName, cMsg )
    LOCAL e := ErrorNew()

    IF EMPTY( cClassName )               // c/b NIL
        cClassName := ""
    ELSEIF !EMPTY( cMsg )
        cClassName += ":"
    END

    IF cMsg == NIL
        cMsg := ""
    END

    e:genCode   := EG_NOMETHOD      // 5.01 also has EG_NOVARMETHOD ("No exported var")
    e:subSystem := "CLASS(Y)"
    e:subCode   := nSubcode
    e:description := cDesc
    e:operation := cClassName + cMsg
    e:severity  := ES_ERROR

    // allow override eg. ::handleErr( e )?
    EVAL( ErrorBlock(), e )
RETURN


METHOD PROCEDURE deferred
    ::error( CSYERR_DEFERRED, "Message " + CsyMsgName( 0 ) + " should be implemented by subclass", ::className )
RETURN

METHOD PROCEDURE protectErr( cMsg )
    ::error( CSYERR_PROTECTED, "Scope violation (protected)", ::className, cMsg )
RETURN

METHOD PROCEDURE hiddenErr( cMsg )
    ::error( CSYERR_HIDDEN, "Scope violation (hidden)", ::className, cMsg )
RETURN

METHOD PROCEDURE msgNotFound( cMsg )
    ::error( CSYERR_NOTFOUND, "Message not found", ::className, cMsg )
RETURN


/*
    readOnlyErr()

    Handles delegated assignment to a read-only variable.  It is executed
    directly, rather than via the kernel's error handler, so is implemented
    a bit differently from other error methods.
*/

METHOD PROCEDURE readOnlyErr()
    ::error( CSYERR_READONLY, ;
        "Illegal assignment to read-only variable in class " + ::className, ;
        nil, MethodName( 1 ) )
RETURN

METHOD PROCEDURE wrongClass( cMsg )
    ::error( CSYERR_WRONGCLASS, "Assigned value is wrong class", ::className, cMsg )
RETURN

METHOD PROCEDURE badMethod( cMsg )
    ::error( CSYERR_BADMETHOD, "Bad or missing method reference", ::className, cMsg )
RETURN


// eof csyobjec.prg
