/*
    csyscala.prg - Class(y) ScalarObject class definition

    Copyright (c) 1992-1994 Anton van Straaten

    Class(y) 2.0 treats standard Clipper data types as though they belong
    to one of the classes Array, Character, Numeric, Logical, Date, Block,
    or Nil.  With the exception of Array, these classes are subclasses of
    the class ScalarObject, defined here.  ScalarObject overrides the
    Object messages 'copy' and 'deepCopy' so that these messages behave
    correctly when sent to scalar objects.  This is essential to the
    correct operation of Object's 'deepCopy' method, which expects all
    values to be able to respond to the 'copy' message.

    The clause "METACLASS Class" in the creation of ScalarObject is not
    necessary, but is more efficient since it tells Class(y) that it
    does not have to create a (provisional) metaclass while creating
    this class.
*/

#include "class(y).ch"


CREATE CLASS ScalarObject METACLASS Class
EXPORT:
    MESSAGE become      METHOD becomeErr    // a scalar cannot "become" another object
    METHOD  copy
    MESSAGE deepCopy    METHOD copy
    METHOD  isScalar
    METHOD  asString
    METHOD  asExpStr

END CLASS


/*
    copy()

    This handles all scalar classes except Array.
    Array is not inherited from the ScalarObject class.
*/

METHOD FUNCTION copy
RETURN self


METHOD isScalar
RETURN .T.


/*
    asString()

    This method is a kludge provided for those people who have defined
    their own scalar classes without an asString() method (most likely
    because they wrote their scalar classes prior to v2.4).  This most
    un-object-oriented and relatively inefficient method provides
    default behavior in such cases.
*/

METHOD asString
    LOCAL cType := VALTYPE( self )

    DO CASE
        CASE cType == 'B'
            RETURN "{ || ... }"
        CASE cType == 'C'
            RETURN self
        CASE cType == 'D'
            RETURN DTOC( self )
        CASE cType == 'L'
            RETURN IF( self, ".T.", ".F." )
        CASE cType == 'N'
            RETURN LTRIM( STR( self ) )
        CASE cType == 'U'
            RETURN "NIL"
    END CASE

RETURN "Error!"


/*
    asExpStr()

    As with asString, this method defines default behavior for asExpStr()
    for systems which do not link in the scalar classes.
*/

METHOD asExpStr
    LOCAL cType := VALTYPE( self )

    IF cType == 'C'
        RETURN ["] + self + ["]
    ELSEIF cType == 'D'
        RETURN [CTOD("] + DTOC( self ) + [")]
    END

RETURN ::asString


#include "csyerror.ch"

METHOD PROCEDURE becomeErr
    ::error(CSYERR_BECOME, "Message 'become' illegally sent to scalar", ::className)
RETURN


// eof csyscala.prg
